#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime

# import interfaces
from pyams_security.interfaces import IPrincipalInfo
from pyams_workflow.interfaces import IWorkflowVersions, IWorkflowManagedContent, VersionError, IWorkflowState, \
    IWorkflowVersion, IWorkflowStateHistoryItem, IWorkflowTransitionEvent, IWorkflowVersionTransitionEvent, IWorkflow
from zope.annotation.interfaces import IAnnotations
from zope.location.interfaces import ISublocations
from zope.traversing.interfaces import ITraversable

# import packages
from persistent import Persistent
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from pyams_utils.adapter import adapter_config, ContextAdapter
from pyams_utils.registry import get_utility
from pyams_utils.request import check_request, query_request
from pyams_utils.traversing import get_parent
from pyramid.events import subscriber
from pyramid.threadlocal import get_current_registry
from zope.container.folder import Folder
from zope.interface import implementer, alsoProvides
from zope.lifecycleevent import ObjectCreatedEvent
from zope.location import locate
from zope.schema.fieldproperty import FieldProperty


@implementer(IWorkflowStateHistoryItem)
class WorkflowHistoryItem(Persistent):
    """Workflow history item"""

    date = FieldProperty(IWorkflowStateHistoryItem['date'])
    source_version = FieldProperty(IWorkflowStateHistoryItem['source_version'])
    source_state = FieldProperty(IWorkflowStateHistoryItem['source_state'])
    target_state = FieldProperty(IWorkflowStateHistoryItem['target_state'])
    transition_id = FieldProperty(IWorkflowStateHistoryItem['transition_id'])
    transition = FieldProperty(IWorkflowStateHistoryItem['transition'])
    principal = FieldProperty(IWorkflowStateHistoryItem['principal'])
    comment = FieldProperty(IWorkflowStateHistoryItem['comment'])

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


@implementer(IWorkflowState)
class WorkflowVersionState(Persistent):
    """Workflow managed content version object"""

    version_id = None
    _state = None
    _state_date = FieldProperty(IWorkflowState['state_date'])
    _state_principal = FieldProperty(IWorkflowState['state_principal'])
    state_urgency = FieldProperty(IWorkflowState['state_urgency'])

    def __init__(self):
        self.history = PersistentList()

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = value
        self._state_date = datetime.utcnow()
        request = query_request()
        if request is not None:
            self._state_principal = request.principal.id

    @property
    def state_date(self):
        return self._state_date

    @property
    def state_principal(self):
        return self._state_principal

    @state_principal.setter
    def state_principal(self, value):
        self._state_principal = value.id if IPrincipalInfo.providedBy(value) else value

    def get_first_state_date(self, states):
        for item in self.history:
            if item.target_state in states:
                return item.date
        return None


@subscriber(IWorkflowTransitionEvent)
def handle_workflow_transition(event):
    """Handle workflow transition"""
    if IWorkflowVersionTransitionEvent.providedBy(event):
        return
    principal = event.principal
    item = WorkflowHistoryItem(date=datetime.utcnow(),
                               source_state=event.source,
                               target_state=event.destination,
                               transition_id=event.transition.transition_id,
                               principal=principal.id if IPrincipalInfo.providedBy(principal) else principal,
                               comment=event.comment)
    IWorkflowState(event.object).history.append(item)


@subscriber(IWorkflowVersionTransitionEvent)
def handle_workflow_version_transition(event):
    """Handle workflow version transition"""
    principal = event.principal
    item = WorkflowHistoryItem(date=datetime.utcnow(),
                               source_version=IWorkflowState(event.old_object).version_id,
                               source_state=event.source,
                               target_state=event.destination,
                               transition_id=event.transition.transition_id,
                               principal=principal.id if IPrincipalInfo.providedBy(principal) else principal,
                               comment=event.comment)
    IWorkflowState(event.object).history.append(item)


WORKFLOW_VERSION_KEY = 'pyams_workflow.version'


@adapter_config(context=IWorkflowVersion, provides=IWorkflowState)
def WorkflowVersionStateFactory(context):
    """Workflow content version state factory"""
    annotations = IAnnotations(context)
    state = annotations.get(WORKFLOW_VERSION_KEY)
    if state is None:
        state = annotations[WORKFLOW_VERSION_KEY] = WorkflowVersionState()
        registry = get_current_registry()
        registry.notify(ObjectCreatedEvent(state))
    return state


@implementer(IWorkflowVersions)
class WorkflowVersions(Folder):
    """Workflow versions adapter"""

    def __init__(self):
        super(WorkflowVersions, self).__init__()
        self.last_version_id = 0
        self.state_by_version = PersistentMapping()
        self.versions_by_state = PersistentMapping()
        self.deleted = PersistentMapping()

    def get_version(self, version_id):
        if version_id is None:
            version_id = self.last_version_id
        try:
            return self[str(version_id)]
        except KeyError:
            raise VersionError("Missing given version ID {0}".format(version_id))

    def get_versions(self, states=None, sort=False, reverse=False):
        if states:
            if isinstance(states, str):
                states = (states, )
            versions = set()
            for state in states:
                if state is None:
                    state = '__none__'
                [versions.add(self[str(version)]) for version in self.versions_by_state.get(state, ())]
            if sort:
                versions = sorted(versions, key=lambda x: int(x.__name__), reverse=reverse)
            return versions
        else:
            return (v for k, v in sorted(self.items(), key=lambda x: int(x[0])))

    def get_last_versions(self, count=1):
        result = list((v for k, v in sorted(self.items(), key=lambda x: -int(x[0]))))
        if count:
            result = result[:count]
        return result

    def add_version(self, content, state, principal=None):
        self.last_version_id += 1
        version_id = self.last_version_id
        # init version state
        alsoProvides(content, IWorkflowVersion)
        wf_state = IWorkflowState(content)
        wf_state.version_id = version_id
        wf_state.state = state
        if principal is not None:
            wf_state.state_principal = principal
        # store new version
        if state is None:
            state = '__none__'
        self[str(version_id)] = content
        self.state_by_version[version_id] = state
        versions = self.versions_by_state.get(state, [])
        versions.append(version_id)
        self.versions_by_state[state] = versions
        return version_id

    def set_state(self, version_id, state, principal=None):
        # update version state
        version = self[str(version_id)]
        wf_state = IWorkflowState(version)
        wf_state.version_id = version_id
        wf_state.state = state
        if principal is not None:
            wf_state.state_principal = principal
        # update versions/states mapping
        if state is None:
            state = '__none__'
        old_state = self.state_by_version[version_id]
        versions = self.versions_by_state[old_state]
        if version_id in versions:
            versions.remove(version_id)
            if versions:
                self.versions_by_state[old_state] = versions
            else:
                del self.versions_by_state[old_state]
        self.state_by_version[version_id] = state
        versions = self.versions_by_state.get(state, [])
        versions.append(version_id)
        self.versions_by_state[state] = versions

    def has_version(self, states):
        if states is None:
            states = '__none__'
        if not isinstance(states, (list, tuple, set)):
            states = {states}
        for state in states:
            if bool(self.versions_by_state.get(state, ())):
                return True
        return False

    def remove_version(self, version_id, state='deleted', comment=None, principal=None):
        if str(version_id) not in self:
            return
        # update version state
        version = self[str(version_id)]
        wf_state = IWorkflowState(version)
        if comment:
            request = check_request()
            translate = request.localizer.translate
            workflow = get_utility(IWorkflow, name=get_parent(self, IWorkflowManagedContent).workflow_name)
            item = WorkflowHistoryItem(date=datetime.utcnow(),
                                       source_version=wf_state.version_id,
                                       source_state=translate(workflow.states.getTerm(wf_state.state).title),
                                       target_state=translate(workflow.states.getTerm(state).title),
                                       principal=request.principal.id,
                                       comment=comment)
            wf_state.history.append(item)
        wf_state.state = state
        if principal is not None:
            wf_state.state_principal = principal
        # remove given version
        state = self.state_by_version[version_id]
        versions = self.versions_by_state[state]
        versions.remove(version_id)
        if versions:
            self.versions_by_state[state] = versions
        else:
            del self.versions_by_state[state]
        del self.state_by_version[version_id]
        self.deleted[version_id] = self[str(version_id)]
        del self[str(version_id)]


WORKFLOW_VERSIONS_KEY = 'pyams_workflow.versions'


@adapter_config(context=IWorkflowManagedContent, provides=IWorkflowVersions)
def WorkflowContentVersionsFactory(context):
    """Workflow versions factory"""
    annotations = IAnnotations(context)
    versions = annotations.get(WORKFLOW_VERSIONS_KEY)
    if versions is None:
        versions = annotations[WORKFLOW_VERSIONS_KEY] = WorkflowVersions()
        registry = get_current_registry()
        registry.notify(ObjectCreatedEvent(versions))
        locate(versions, context, '++versions++')
    return versions


@adapter_config(context=IWorkflowVersion, provides=IWorkflowVersions)
def WorkflowVersionVersionsFactory(context):
    """Workflow versions factory for version"""
    parent = get_parent(context, IWorkflowManagedContent)
    if parent is not None:
        return IWorkflowVersions(parent)


@adapter_config(name='versions', context=IWorkflowManagedContent, provides=ITraversable)
class WorkflowVersionsTraverser(ContextAdapter):
    """++versions++ namespace traverser"""

    def traverse(self, name, furtherpath=None):
        versions = IWorkflowVersions(self.context)
        if name:
            return versions[name]
        else:
            return versions


@adapter_config(name='versions', context=IWorkflowManagedContent, provides=ISublocations)
class WorkflowVersionsSublocations(ContextAdapter):
    """Workflow versions sub-locations"""

    def sublocations(self):
        return IWorkflowVersions(self.context).values()
